"""
Tests for ways the helper can be used.
"""

import requests
import requests_mock
import responses
from flask import Flask

from requests_mock_flask import add_flask_app_to_mock


class TestResponses:
    """
    Tests for using the helper with ``responses``.
    """

    def test_context_manager(self) -> None:
        """
        It is possible to use the helper with a ``responses`` context manager.
        """
        app = Flask(__name__)

        @app.route('/')
        def _() -> str:
            return 'Hello, World!'

        with responses.RequestsMock(
            assert_all_requests_are_fired=False,
        ) as resp_m:
            add_flask_app_to_mock(
                mock_obj=resp_m,
                flask_app=app,
                base_url='http://www.example.com',
            )

            response = requests.get('http://www.example.com')

        assert response.status_code == 200
        assert response.text == 'Hello, World!'

    @responses.activate
    def test_decorator(self) -> None:
        """
        It is possible to use the helper with a ``responses`` decorator.
        """
        app = Flask(__name__)

        @app.route('/')
        def _() -> str:
            return 'Hello, World!'

        add_flask_app_to_mock(
            mock_obj=responses,
            flask_app=app,
            base_url='http://www.example.com',
        )

        response = requests.get('http://www.example.com')

        assert response.status_code == 200
        assert response.text == 'Hello, World!'


class TestRequestsMock:
    """
    Tests for using the helper with ``requests_mock``.
    """

    def test_context_manager(self) -> None:
        """
        It is possible to use the helper with a ``requests_mock`` context
        manager.
        """
        app = Flask(__name__)

        @app.route('/')
        def _() -> str:
            return 'Hello, World!'

        with requests_mock.Mocker() as resp_m:
            add_flask_app_to_mock(
                mock_obj=resp_m,
                flask_app=app,
                base_url='http://www.example.com',
            )

            response = requests.get('http://www.example.com')

        assert response.status_code == 200
        assert response.text == 'Hello, World!'

    def test_fixture(  # pylint: disable=redefined-outer-name
        self,
        requests_mock: requests_mock.Mocker,
    ) -> None:
        """
        It is possible to use the helper with a ``requests_mock`` fixture.
        """
        app = Flask(__name__)

        @app.route('/')
        def _() -> str:
            return 'Hello, World!'

        add_flask_app_to_mock(
            mock_obj=requests_mock,
            flask_app=app,
            base_url='http://www.example.com',
        )

        response = requests.get('http://www.example.com')

        assert response.status_code == 200

    def test_adapter(self) -> None:
        """
        It is possible to use the helper with a ``requests_mock`` fixture.
        """
        app = Flask(__name__)

        @app.route('/')
        def _() -> str:
            return 'Hello, World!'

        session = requests.Session()
        adapter = requests_mock.Adapter()
        session.mount('mock', adapter)

        add_flask_app_to_mock(
            mock_obj=adapter,
            flask_app=app,
            base_url='mock://www.example.com',
        )

        response = session.get('mock://www.example.com')

        assert response.status_code == 200
        assert response.text == 'Hello, World!'



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Configuration for Sphinx.
"""

from pkg_resources import get_distribution

# pylint: disable=invalid-name

# -- General configuration ------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.napoleon',
    'sphinx_autodoc_typehints',
    'sphinxcontrib.spelling',
    'sphinx-prompt',
    'sphinx_substitution_extensions',
]

project = 'requests-mock-flask'

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# General information about the project.
project = 'Requests-Mock-Flask'
copyright = '2020, Adam Dangoor'  # pylint: disable=redefined-builtin
author = 'Adam Dangoor'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# Use ``pkg_resources`` as per
# https://github.com/pypa/setuptools_scm#usage-from-sphinx.
version = get_distribution(project).version
_month, _day, _year, *_ = version.split('.')
release = f'{_month}.{_day}.{_year}'

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False

# -- Options for HTML output ----------------------------------------------

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
html_show_sphinx = False

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
html_show_copyright = False

# Output file base name for HTML help builder.
htmlhelp_basename = 'requestsmockflaskdoc'

spelling_word_list_filename = '../../spelling_private_dict.txt'

rst_prolog = f"""
.. |project| replace:: {project}
.. |release| replace:: {release}
.. |github-owner| replace:: adamtheturtle
.. |github-repository| replace:: requests-mock-flask
"""

always_document_param_types = True



"""
Package for ``requests_mock_flask``.
"""

import re
from functools import partial
from typing import Any, Dict, Optional, Tuple, Union
from urllib.parse import urljoin

import werkzeug
from flask import Flask
from requests import PreparedRequest
from requests_mock.request import _RequestObjectProxy
from requests_mock.response import _Context
from werkzeug.http import parse_cookie


def add_flask_app_to_mock(
    # We use the ``Any`` type to support:
    # ``requests_mock.Mocker, ``responses.RequestsMock`` and the ``responses``
    # module.
    mock_obj: Any,
    flask_app: Flask,
    base_url: str,
) -> None:
    """
    Make it so that requests sent to the ``base_url`` are forwarded to the
    ``Flask`` app, when in the context of the ``mock_obj``.
    """
    if hasattr(mock_obj, 'add_callback'):
        resp_callback = partial(_responses_callback, flask_app=flask_app)
        register_method = partial(
            mock_obj.add_callback,
            callback=resp_callback,
        )
    else:
        req_m_callback = partial(_requests_mock_callback, flask_app=flask_app)
        register_method = partial(mock_obj.register_uri, text=req_m_callback)

    for rule in flask_app.url_map.iter_rules():
        # We replace everything inside angle brackets with a match for any
        # string of characters of length > 0.
        path_to_match = re.sub(pattern='<.+>', repl='.+', string=rule.rule)
        pattern = urljoin(base_url, path_to_match)
        url = re.compile(pattern)

        for method in rule.methods:
            register_method(method=method, url=url)


def _responses_callback(
    request: PreparedRequest,
    flask_app: Flask,
) -> Tuple[int, Dict[str, Optional[Union[str, int, bool]]], bytes]:
    """
    Given a request to the flask app, send an equivalent request to an in
    memory fake of the flask app and return some key details of the
    response.

    :param request: The incoming request to pass onto the flask app.
    :param flask_app: The Flask application to pass requests to.
    :return: A tuple of status code, response headers and response data
        from the flask app.
    """
    test_client = flask_app.test_client()
    # See parameters at
    # https://werkzeug.palletsprojects.com/en/0.15.x/test/#werkzeug.test.EnvironBuilder
    cookie_string = request.headers.get('Cookie', '')
    cookie_list = cookie_string.split(';')
    cookie_list_no_empty = [item for item in cookie_list if item]
    request_cookies = [
        list(parse_cookie(cookie).items())[0]
        for cookie in cookie_list_no_empty
    ]
    cookies_dict = dict(request_cookies)

    for key, value in cookies_dict.items():
        test_client.set_cookie(
            server_name='',
            key=key,
            value=value,
        )

    environ_builder = werkzeug.test.EnvironBuilder(
        path=request.path_url,
        method=str(request.method),
        data=request.body,
        headers=dict(request.headers),
    )
    environ = environ_builder.get_environ()  # type: ignore
    if 'Content-Length' in request.headers:
        environ['CONTENT_LENGTH'] = request.headers['Content-Length']
        environ['wsgi.input_terminated'] = True
    response = test_client.open(environ)

    result = (response.status_code, dict(response.headers), response.data)
    return result


def _requests_mock_callback(
    request: _RequestObjectProxy,
    context: _Context,
    flask_app: Flask,
) -> str:
    """
    Given a request to the flask app, send an equivalent request to an in
    memory fake of the flask app and return some key details of the
    response.

    :param request: The incoming request to pass onto the flask app.
    :param context: An object containing the collected known data about this
        response.
    :param flask_app: The Flask application to pass requests to.
    :return: A tuple of status code, response headers and response data
        from the flask app.
    """
    test_client = flask_app.test_client()
    # See parameters at
    # https://werkzeug.palletsprojects.com/en/0.15.x/test/#werkzeug.test.EnvironBuilder
    cookie_string = request.headers.get('Cookie', '')
    cookie_list = cookie_string.split(';')
    cookie_list_no_empty = [item for item in cookie_list if item]
    request_cookies = [
        list(parse_cookie(cookie).items())[0]
        for cookie in cookie_list_no_empty
    ]
    cookies_dict = dict(request_cookies)

    for key, value in cookies_dict.items():
        test_client.set_cookie(
            server_name='',
            key=key,
            value=value,
        )

    environ_builder = werkzeug.test.EnvironBuilder(
        path=request.path_url,
        method=request.method,
        headers=dict(request.headers),
        data=request.body,
    )
    environ = environ_builder.get_environ()  # type: ignore
    if 'Content-Length' in request.headers:
        environ['CONTENT_LENGTH'] = request.headers['Content-Length']
        environ['wsgi.input_terminated'] = True
    response = test_client.open(environ)

    context.headers = response.headers
    context.status_code = response.status_code
    return str(response.data.decode())



# pylint: disable=too-many-lines
"""
Tests for the ``requests_mock_flask`` package.

Test with a bunch of route types as per:
https://flask.palletsprojects.com/en/1.1.x/quickstart/#variable-rules
"""

import json
import uuid
from typing import Tuple

import pytest
import requests
import requests_mock
import responses
import werkzeug
from flask import Flask, Response, jsonify, make_response, request
from flask_negotiate import consumes

from requests_mock_flask import add_flask_app_to_mock


def test_simple_route() -> None:
    """
    A simple GET route works.
    """
    app = Flask(__name__)

    @app.route('/')
    def _() -> str:
        return 'Hello, World!'

    test_client = app.test_client()
    response = test_client.get('/')

    expected_status_code = 200
    expected_content_type = 'text/html; charset=utf-8'
    expected_data = b'Hello, World!'

    assert response.status_code == expected_status_code
    assert response.headers['Content-Type'] == expected_content_type
    assert response.data == expected_data

    with responses.RequestsMock(assert_all_requests_are_fired=False) as resp_m:
        add_flask_app_to_mock(
            mock_obj=resp_m,
            flask_app=app,
            base_url='http://www.example.com',
        )

        responses_response = requests.get('http://www.example.com')

    assert responses_response.status_code == expected_status_code
    assert responses_response.headers['Content-Type'] == expected_content_type
    assert responses_response.text == expected_data.decode()

    with requests_mock.Mocker() as resp_m:
        add_flask_app_to_mock(
            mock_obj=resp_m,
            flask_app=app,
            base_url='http://www.example.com',
        )

        req_mock_response = requests.get('http://www.example.com')

    assert req_mock_response.status_code == expected_status_code
    assert req_mock_response.headers['Content-Type'] == expected_content_type
    assert req_mock_response.text == expected_data.decode()


def test_headers() -> None:
    """
    Request headers are sent.
    """
    app = Flask(__name__)

    @app.route('/')
    def _() -> str:
        assert 'Content-Type' not in request.headers
        assert request.headers['hello'] == 'world'
        return 'Hello, World!'

    test_client = app.test_client()
    response = test_client.get('/', headers={'hello': 'world'})

    expected_status_code = 200
    expected_content_type = 'text/html; charset=utf-8'
    expected_data = b'Hello, World!'

    assert response.status_code == expected_status_code
    assert response.headers['Content-Type'] == expected_content_type
    assert response.data == expected_data

    with responses.RequestsMock(assert_all_requests_are_fired=False) as resp_m:
        add_flask_app_to_mock(
            mock_obj=resp_m,
            flask_app=app,
            base_url='http://www.example.com',
        )

        responses_response = requests.get(
            'http://www.example.com',
            headers={'hello': 'world'},
        )

    assert responses_response.status_code == expected_status_code
    assert responses_response.headers['Content-Type'] == expected_content_type
    assert responses_response.text == expected_data.decode()

    with requests_mock.Mocker() as resp_m:
        add_flask_app_to_mock(
            mock_obj=resp_m,
            flask_app=app,
            base_url='http://www.example.com',
        )

        req_mock_response = requests.get(
            'http://www.example.com',
            headers={'hello': 'world'},
        )

    assert req_mock_response.status_code == expected_status_code
    assert req_mock_response.headers['Content-Type'] == expected_content_type
    assert req_mock_response.text == expected_data.decode()


def test_route_with_json() -> None:
    """
    A route that returns JSON data works.
    """
    app = Flask(__name__)

    @app.route('/')
    def _() -> Tuple[Response, int]:
        return jsonify({'hello': 'world'}), 201

    test_client = app.test_client()
    response = test_client.get('/')

    expected_status_code = 201
    expected_content_type = 'application/json'
    expected_json = {'hello': 'world'}

    assert response.status_code == expected_status_code
    assert response.headers['Content-Type'] == expected_content_type
    assert response.json == expected_json

    with responses.RequestsMock(assert_all_requests_are_fired=False) as resp_m:
        add_flask_app_to_mock(
            mock_obj=resp_m,
            flask_app=app,
            base_url='http://www.example.com',
        )

        responses_response = requests.get('http://www.example.com')

    assert responses_response.status_code == expected_status_code
    assert responses_response.headers['Content-Type'] == expected_content_type
    assert responses_response.json() == expected_json

    with requests_mock.Mocker() as resp_m:
        add_flask_app_to_mock(
            mock_obj=resp_m,
            flask_app=app,
            base_url='http://www.example.com',
        )

        req_mock_response = requests.get('http://www.example.com')

    assert req_mock_response.status_code == expected_status_code
    assert req_mock_response.headers['Content-Type'] == expected_content_type
    assert req_mock_response.json() == expected_json


def test_route_with_variable_no_type_given() -> None:
    """
    A route with a variable works.
    """
    app = Flask(__name__)

    @app.route('/<my_variable>')
    def _(my_variable: str) -> str:
        return 'Hello: ' + my_variable

    test_client = app.test_client()
    response = test_client.get('/Frasier')

    expected_status_code = 200
    expected_content_type = 'text/html; charset=utf-8'
    expected_data = b'Hello: Frasier'

    assert response.status_code == expected_status_code
    assert response.headers['Content-Type'] == expected_content_type
    assert response.data == expected_data

    with responses.RequestsMock(assert_all_requests_are_fired=False) as resp_m:
        add_flask_app_to_mock(
            mock_obj=resp_m,
            flask_app=app,
            base_url='http://www.example.com',
        )

        responses_response = requests.get('http://www.example.com/Frasier')

    assert responses_response.status_code == expected_status_code
    assert responses_response.headers['Content-Type'] == expected_content_type
    assert responses_response.text == expected_data.decode()

    with requests_mock.Mocker() as resp_m:
        add_flask_app_to_mock(
            mock_obj=resp_m,
            flask_app=app,
            base_url='http://www.example.com',
        )

        req_mock_response = requests.get('http://www.example.com/Frasier')

    assert req_mock_response.status_code == expected_status_code
    assert req_mock_response.headers['Content-Type'] == expected_content_type
    assert req_mock_response.text == expected_data.decode()


def test_route_with_string_variable() -> None:
    """
    A route with a string variable works.
    """
    app = Flask(__name__)

    @app.route('/<string:my_variable>')
    def _(my_variable: str) -> str:
        return 'Hello: ' + my_variable

    test_client = app.test_client()
    response = test_client.get('/Frasier')

    expected_status_code = 200
    expected_content_type = 'text/html; charset=utf-8'
    expected_data = b'Hello: Frasier'

    assert response.status_code == expected_status_code
    assert response.headers['Content-Type'] == expected_content_type
    assert response.data == expected_data

    with responses.RequestsMock(assert_all_requests_are_fired=False) as resp_m:
        add_flask_app_to_mock(
            mock_obj=resp_m,
            flask_app=app,
            base_url='http://www.example.com',
        )

        responses_response = requests.get('http://www.example.com/Frasier')

    assert responses_response.status_code == expected_status_code
    assert responses_response.headers['Content-Type'] == expected_content_type
    assert responses_response.text == expected_data.decode()

    with requests_mock.Mocker() as resp_m:
        add_flask_app_to_mock(
            mock_obj=resp_m,
            flask_app=app,
            base_url='http://www.example.com',
        )

        req_mock_response = requests.get('http://www.example.com/Frasier')

    assert req_mock_response.status_code == expected_status_code
    assert req_mock_response.headers['Content-Type'] == expected_content_type
    assert req_mock_response.text == expected_data.decode()


def test_route_with_int_variable() -> None:
    """
    A route with an int variable works.
    """
    app = Flask(__name__)

    @app.route('/<int:my_variable>')
    def _(my_variable: int) -> str:
        return 'Hello: ' + str(my_variable + 5)

    test_client = app.test_client()
    response = test_client.get('/4')

    expected_status_code = 200
    expected_content_type = 'text/html; charset=utf-8'
    expected_data = b'Hello: 9'

    assert response.status_code == expected_status_code
    assert response.headers['Content-Type'] == expected_content_type
    assert response.data == expected_data

    with responses.RequestsMock(assert_all_requests_are_fired=False) as resp_m:
        add_flask_app_to_mock(
            mock_obj=resp_m,
            flask_app=app,
            base_url='http://www.example.com',
        )

        responses_response = requests.get('http://www.example.com/4')

    assert responses_response.status_code == expected_status_code
    assert responses_response.headers['Content-Type'] == expected_content_type
    assert responses_response.text == expected_data.decode()

    with requests_mock.Mocker() as resp_m:
        add_flask_app_to_mock(
            mock_obj=resp_m,
            flask_app=app,
            base_url='http://www.example.com',
        )

        req_mock_response = requests.get('http://www.example.com/4')

    assert req_mock_response.status_code == expected_status_code
    assert req_mock_response.headers['Content-Type'] == expected_content_type
    assert req_mock_response.text == expected_data.decode()


def test_route_with_float_variable() -> None:
    """
    A route with a float variable works.
    """
    app = Flask(__name__)

    @app.route('/<float:my_variable>')
    def _(my_variable: float) -> str:
        return 'Hello: ' + str(my_variable + 5)

    test_client = app.test_client()
    response = test_client.get('/4.0')

    expected_status_code = 200
    expected_content_type = 'text/html; charset=utf-8'
    expected_data = b'Hello: 9.0'

    assert response.status_code == expected_status_code
    assert response.headers['Content-Type'] == expected_content_type
    assert response.data == expected_data

    with responses.RequestsMock(assert_all_requests_are_fired=False) as resp_m:
        add_flask_app_to_mock(
            mock_obj=resp_m,
            flask_app=app,
            base_url='http://www.example.com',
        )

        responses_response = requests.get('http://www.example.com/4.0')

    assert responses_response.status_code == expected_status_code
    assert responses_response.headers['Content-Type'] == expected_content_type
    assert responses_response.text == expected_data.decode()

    with requests_mock.Mocker() as resp_m:
        add_flask_app_to_mock(
            mock_obj=resp_m,
            flask_app=app,
            base_url='http://www.example.com',
        )

        req_mock_response = requests.get('http://www.example.com/4.0')

    assert req_mock_response.status_code == expected_status_code
    assert req_mock_response.headers['Content-Type'] == expected_content_type
    assert req_mock_response.text == expected_data.decode()


def test_route_with_path_variable_with_slash() -> None:
    """
    A route with a path variable works.
    """
    app = Flask(__name__)

    @app.route('/<path:my_variable>')
    def _(my_variable: str) -> str:
        return 'Hello: ' + my_variable

    test_client = app.test_client()
    response = test_client.get('/foo/bar')

    expected_status_code = 200
    expected_content_type = 'text/html; charset=utf-8'
    expected_data = b'Hello: foo/bar'

    assert response.status_code == expected_status_code
    assert response.headers['Content-Type'] == expected_content_type
    assert response.data == expected_data

    with responses.RequestsMock(assert_all_requests_are_fired=False) as resp_m:
        add_flask_app_to_mock(
            mock_obj=resp_m,
            flask_app=app,
            base_url='http://www.example.com',
        )

        responses_response = requests.get('http://www.example.com/foo/bar')

    assert responses_response.status_code == expected_status_code
    assert responses_response.headers['Content-Type'] == expected_content_type
    assert responses_response.text == expected_data.decode()

    with requests_mock.Mocker() as resp_m:
        add_flask_app_to_mock(
            mock_obj=resp_m,
            flask_app=app,
            base_url='http://www.example.com',
        )

        req_mock_response = requests.get('http://www.example.com/foo/bar')

    assert req_mock_response.status_code == expected_status_code
    assert req_mock_response.headers['Content-Type'] == expected_content_type
    assert req_mock_response.text == expected_data.decode()


def test_route_with_string_variable_with_slash() -> None:
    """
    A route with a string variable when given a slash works.
    """
    app = Flask(__name__)

    @app.route('/<string:my_variable>')
    def _(_: str) -> str:
        ...  # pragma: no cover

    test_client = app.test_client()
    response = test_client.get('/foo/bar')

    expected_status_code = 404
    expected_content_type = 'text/html; charset=utf-8'

    assert response.status_code == expected_status_code
    assert response.headers['Content-Type'] == expected_content_type
    assert b'not found on the server' in response.data

    with responses.RequestsMock(assert_all_requests_are_fired=False) as resp_m:
        add_flask_app_to_mock(
            mock_obj=resp_m,
            flask_app=app,
            base_url='http://www.example.com',
        )

        responses_response = requests.get('http://www.example.com/foo/bar')

    assert responses_response.status_code == expected_status_code
    assert responses_response.headers['Content-Type'] == expected_content_type
    assert 'not found on the server' in responses_response.text

    with requests_mock.Mocker() as resp_m:
        add_flask_app_to_mock(
            mock_obj=resp_m,
            flask_app=app,
            base_url='http://www.example.com',
        )

        req_mock_response = requests.get('http://www.example.com/foo/bar')

    assert req_mock_response.status_code == expected_status_code
    assert req_mock_response.headers['Content-Type'] == expected_content_type
    assert 'not found on the server' in req_mock_response.text


def test_route_with_uuid_variable() -> None:
    """
    A route with a uuid variable works.
    """
    app = Flask(__name__)

    @app.route('/<uuid:my_variable>')
    def _(my_variable: uuid.UUID) -> str:
        return 'Hello: ' + my_variable.hex

    test_client = app.test_client()
    random_uuid = uuid.uuid4()
    response = test_client.get(f'/{random_uuid}')

    expected_status_code = 200
    expected_content_type = 'text/html; charset=utf-8'
    expected_data = bytes(f'Hello: {random_uuid.hex}', 'utf-8')

    assert response.status_code == expected_status_code
    assert response.headers['Content-Type'] == expected_content_type
    assert response.data == expected_data

    with responses.RequestsMock(assert_all_requests_are_fired=False) as resp_m:
        add_flask_app_to_mock(
            mock_obj=resp_m,
            flask_app=app,
            base_url='http://www.example.com',
        )

        responses_response = requests.get(
            f'http://www.example.com/{random_uuid}',
        )

    assert responses_response.status_code == expected_status_code
    assert responses_response.headers['Content-Type'] == expected_content_type
    assert responses_response.text == expected_data.decode()

    with requests_mock.Mocker() as resp_m:
        add_flask_app_to_mock(
            mock_obj=resp_m,
            flask_app=app,
            base_url='http://www.example.com',
        )

        req_mock_response = requests.get(
            f'http://www.example.com/{random_uuid}',
        )

    assert req_mock_response.status_code == expected_status_code
    assert req_mock_response.headers['Content-Type'] == expected_content_type
    assert req_mock_response.text == expected_data.decode()


def test_nested_path() -> None:
    """
    A route with a variable nested in a path works.
    """
    app = Flask(__name__)

    @app.route('/users/<int:my_variable>/posts')
    def _(my_variable: int) -> str:
        return 'Posts for: ' + str(my_variable)

    test_client = app.test_client()
    response = test_client.get('/users/4/posts')

    expected_status_code = 200
    expected_content_type = 'text/html; charset=utf-8'
    expected_data = b'Posts for: 4'

    assert response.status_code == expected_status_code
    assert response.headers['Content-Type'] == expected_content_type
    assert response.data == expected_data

    with responses.RequestsMock(assert_all_requests_are_fired=False) as resp_m:
        add_flask_app_to_mock(
            mock_obj=resp_m,
            flask_app=app,
            base_url='http://www.example.com',
        )

        responses_response = requests.get(
            'http://www.example.com/users/4/posts',
        )

    assert responses_response.status_code == expected_status_code
    assert responses_response.headers['Content-Type'] == expected_content_type
    assert responses_response.text == expected_data.decode()

    with requests_mock.Mocker() as resp_m:
        add_flask_app_to_mock(
            mock_obj=resp_m,
            flask_app=app,
            base_url='http://www.example.com',
        )

        req_mock_response = requests.get(
            'http://www.example.com/users/4/posts',
        )

    assert req_mock_response.status_code == expected_status_code
    assert req_mock_response.headers['Content-Type'] == expected_content_type
    assert req_mock_response.text == expected_data.decode()


def test_route_with_multiple_variables() -> None:
    """
    A route with multiple variables works.
    """
    app = Flask(__name__)

    @app.route('/users/<string:my_org>/<string:my_user>/posts')
    def _(my_org: str, my_user: str) -> str:
        return 'Posts for: ' + my_org + '/' + my_user

    test_client = app.test_client()
    response = test_client.get('/users/cranes/frasier/posts')

    expected_status_code = 200
    expected_content_type = 'text/html; charset=utf-8'
    expected_data = b'Posts for: cranes/frasier'

    assert response.status_code == expected_status_code
    assert response.headers['Content-Type'] == expected_content_type
    assert response.data == expected_data

    with responses.RequestsMock(assert_all_requests_are_fired=False) as resp_m:
        add_flask_app_to_mock(
            mock_obj=resp_m,
            flask_app=app,
            base_url='http://www.example.com',
        )

        responses_response = requests.get(
            'http://www.example.com/users/cranes/frasier/posts',
        )

    assert responses_response.status_code == expected_status_code
    assert responses_response.headers['Content-Type'] == expected_content_type
    assert responses_response.text == expected_data.decode()

    with requests_mock.Mocker() as resp_m:
        add_flask_app_to_mock(
            mock_obj=resp_m,
            flask_app=app,
            base_url='http://www.example.com',
        )

        req_mock_response = requests.get(
            'http://www.example.com/users/cranes/frasier/posts',
        )

    assert req_mock_response.status_code == expected_status_code
    assert req_mock_response.headers['Content-Type'] == expected_content_type
    assert req_mock_response.text == expected_data.decode()


def test_post_verb() -> None:
    """
    A route with the POST verb works.
    """
    app = Flask(__name__)

    @app.route('/', methods=['POST'])
    def _() -> str:
        return 'Hello, World!'

    test_client = app.test_client()
    response = test_client.post('/')

    expected_status_code = 200
    expected_content_type = 'text/html; charset=utf-8'
    expected_data = b'Hello, World!'

    assert response.status_code == expected_status_code
    assert response.headers['Content-Type'] == expected_content_type
    assert response.data == expected_data

    with responses.RequestsMock(assert_all_requests_are_fired=False) as resp_m:
        add_flask_app_to_mock(
            mock_obj=resp_m,
            flask_app=app,
            base_url='http://www.example.com',
        )

        responses_response = requests.post('http://www.example.com/')

    assert responses_response.status_code == expected_status_code
    assert responses_response.headers['Content-Type'] == expected_content_type
    assert responses_response.text == expected_data.decode()

    with requests_mock.Mocker() as resp_m:
        add_flask_app_to_mock(
            mock_obj=resp_m,
            flask_app=app,
            base_url='http://www.example.com',
        )

        req_mock_response = requests.post('http://www.example.com/')

    assert req_mock_response.status_code == expected_status_code
    assert req_mock_response.headers['Content-Type'] == expected_content_type
    assert req_mock_response.text == expected_data.decode()


@pytest.mark.parametrize(
    'custom_content_length',
    ['1', '100'],
)
def test_incorrect_content_length(custom_content_length: str) -> None:
    """
    Custom content length headers are passed through to the Flask endpoint.
    """
    app = Flask(__name__)
    data = b'12345'

    @app.route('/', methods=['POST'])
    def _() -> str:
        assert len(data) == len(request.data)
        assert request.headers['Content-Length'] == custom_content_length
        return ''

    test_client = app.test_client()
    environ_builder = werkzeug.test.EnvironBuilder(
        path='/',
        method='POST',
        data=b'12345',
    )
    environ = environ_builder.get_environ()  # type: ignore
    environ['CONTENT_LENGTH'] = custom_content_length
    environ['wsgi.input_terminated'] = True
    response = test_client.open(environ)

    expected_status_code = 200

    assert response.status_code == expected_status_code

    requests_request = requests.Request(
        method='POST',
        url='http://www.example.com/',
        data=data,
    ).prepare()
    requests_request.headers['Content-Length'] = custom_content_length

    with responses.RequestsMock(assert_all_requests_are_fired=False) as resp_m:
        add_flask_app_to_mock(
            mock_obj=resp_m,
            flask_app=app,
            base_url='http://www.example.com',
        )

        session = requests.Session()
        responses_response = session.send(  # type: ignore
            request=requests_request,
        )

    assert responses_response.status_code == expected_status_code

    with requests_mock.Mocker() as resp_m:
        add_flask_app_to_mock(
            mock_obj=resp_m,
            flask_app=app,
            base_url='http://www.example.com',
        )

        session = requests.Session()
        req_mock_response = session.send(  # type: ignore
            request=requests_request,
        )

    assert req_mock_response.status_code == expected_status_code


def test_multiple_http_verbs() -> None:
    """
    A route with multiple verbs works.
    """
    app = Flask(__name__)

    @app.route('/', methods=['GET', 'POST'])
    def _() -> str:
        return 'Hello, World!'

    test_client = app.test_client()
    get_response = test_client.get('/')
    post_response = test_client.post('/')

    expected_status_code = 200
    expected_content_type = 'text/html; charset=utf-8'
    expected_data = b'Hello, World!'

    assert get_response.status_code == expected_status_code
    assert get_response.headers['Content-Type'] == expected_content_type
    assert get_response.data == expected_data

    assert post_response.status_code == expected_status_code
    assert post_response.headers['Content-Type'] == expected_content_type
    assert post_response.data == expected_data

    with responses.RequestsMock(assert_all_requests_are_fired=False) as resp_m:
        add_flask_app_to_mock(
            mock_obj=resp_m,
            flask_app=app,
            base_url='http://www.example.com',
        )

        responses_get_response = requests.get('http://www.example.com/')
        responses_post_response = requests.post('http://www.example.com/')

    assert responses_get_response.status_code == expected_status_code
    assert (
        responses_get_response.headers['Content-Type'] == expected_content_type
    )
    assert responses_get_response.text == expected_data.decode()

    assert responses_post_response.status_code == expected_status_code
    assert (
        responses_post_response.headers['Content-Type']
        == expected_content_type
    )
    assert responses_post_response.text == expected_data.decode()

    with requests_mock.Mocker() as resp_m:
        add_flask_app_to_mock(
            mock_obj=resp_m,
            flask_app=app,
            base_url='http://www.example.com',
        )

        req_mock_get_response = requests.get('http://www.example.com/')
        req_mock_post_response = requests.post('http://www.example.com/')

    assert req_mock_get_response.status_code == expected_status_code
    assert (
        req_mock_get_response.headers['Content-Type'] == expected_content_type
    )
    assert req_mock_get_response.text == expected_data.decode()

    assert req_mock_post_response.status_code == expected_status_code
    assert (
        req_mock_post_response.headers['Content-Type'] == expected_content_type
    )
    assert req_mock_post_response.text == expected_data.decode()


def test_wrong_type_given() -> None:
    """
    A route with the wrong type given works.
    """
    app = Flask(__name__)

    @app.route('/<int:my_variable>')
    def _(_: int) -> str:
        ...  # pragma: no cover

    test_client = app.test_client()
    response = test_client.get('/a')

    expected_status_code = 404
    expected_content_type = 'text/html; charset=utf-8'

    assert response.status_code == expected_status_code
    assert response.headers['Content-Type'] == expected_content_type
    assert b'not found on the server' in response.data

    with responses.RequestsMock(assert_all_requests_are_fired=False) as resp_m:
        add_flask_app_to_mock(
            mock_obj=resp_m,
            flask_app=app,
            base_url='http://www.example.com',
        )

        responses_response = requests.get('http://www.example.com/a')

    assert responses_response.status_code == expected_status_code
    assert responses_response.headers['Content-Type'] == expected_content_type
    assert 'not found on the server' in responses_response.text

    with requests_mock.Mocker() as resp_m:
        add_flask_app_to_mock(
            mock_obj=resp_m,
            flask_app=app,
            base_url='http://www.example.com',
        )

        req_mock_response = requests.get('http://www.example.com/a')

    assert req_mock_response.status_code == expected_status_code
    assert req_mock_response.headers['Content-Type'] == expected_content_type
    assert 'not found on the server' in req_mock_response.text


def test_404_no_such_method() -> None:
    """
    A route with the wrong method given works.
    """
    app = Flask(__name__)

    @app.route('/')
    def _() -> str:
        ...  # pragma: no cover

    test_client = app.test_client()
    response = test_client.post('/')

    expected_status_code = 405
    expected_content_type = 'text/html; charset=utf-8'

    assert response.status_code == expected_status_code
    assert response.headers['Content-Type'] == expected_content_type
    assert b'not allowed for the requested URL.' in response.data

    with responses.RequestsMock(assert_all_requests_are_fired=False) as resp_m:
        add_flask_app_to_mock(
            mock_obj=resp_m,
            flask_app=app,
            base_url='http://www.example.com',
        )

        with pytest.raises(requests.exceptions.ConnectionError):
            requests.post('http://www.example.com/')

    with requests_mock.Mocker() as resp_m:
        add_flask_app_to_mock(
            mock_obj=resp_m,
            flask_app=app,
            base_url='http://www.example.com',
        )

        with pytest.raises(requests_mock.exceptions.NoMockAddress):
            requests.post('http://www.example.com/')


def test_request_needs_content_type() -> None:
    """
    Routes which require a content type are supported.
    """
    app = Flask(__name__)

    @app.route('/')
    @consumes('application/json')
    def _() -> str:
        return 'Hello, World!'

    test_client = app.test_client()
    response = test_client.get('/', content_type='application/json')

    expected_status_code = 200
    expected_content_type = 'text/html; charset=utf-8'
    expected_data = b'Hello, World!'

    assert response.status_code == expected_status_code
    assert response.headers['Content-Type'] == expected_content_type
    assert response.data == expected_data

    with responses.RequestsMock(assert_all_requests_are_fired=False) as resp_m:
        add_flask_app_to_mock(
            mock_obj=resp_m,
            flask_app=app,
            base_url='http://www.example.com',
        )

        responses_response = requests.get(
            'http://www.example.com',
            headers={'Content-Type': 'application/json'},
        )

    assert responses_response.status_code == expected_status_code
    assert responses_response.headers['Content-Type'] == expected_content_type
    assert responses_response.text == expected_data.decode()

    with requests_mock.Mocker() as resp_m:
        add_flask_app_to_mock(
            mock_obj=resp_m,
            flask_app=app,
            base_url='http://www.example.com',
        )

        req_mock_response = requests.get(
            'http://www.example.com',
            headers={'Content-Type': 'application/json'},
        )

    assert req_mock_response.status_code == expected_status_code
    assert req_mock_response.headers['Content-Type'] == expected_content_type
    assert req_mock_response.text == expected_data.decode()


def test_request_needs_data() -> None:
    """
    Routes which require data are supported.
    """
    app = Flask(__name__)

    @app.route('/')
    @consumes('application/json')
    def _() -> str:
        return str(request.get_json()['hello'])

    test_client = app.test_client()
    response = test_client.get(
        '/',
        content_type='application/json',
        data=json.dumps({'hello': 'world'}),
    )

    expected_status_code = 200
    expected_content_type = 'text/html; charset=utf-8'
    expected_data = b'world'

    assert response.status_code == expected_status_code
    assert response.headers['Content-Type'] == expected_content_type
    assert response.data == expected_data

    with responses.RequestsMock(assert_all_requests_are_fired=False) as resp_m:
        add_flask_app_to_mock(
            mock_obj=resp_m,
            flask_app=app,
            base_url='http://www.example.com',
        )

        responses_response = requests.get(
            'http://www.example.com',
            headers={'Content-Type': 'application/json'},
            data=json.dumps({'hello': 'world'}),
        )

    assert responses_response.status_code == expected_status_code
    assert responses_response.headers['Content-Type'] == expected_content_type
    assert responses_response.text == expected_data.decode()

    with requests_mock.Mocker() as resp_m:
        add_flask_app_to_mock(
            mock_obj=resp_m,
            flask_app=app,
            base_url='http://www.example.com',
        )

        req_mock_response = requests.get(
            'http://www.example.com',
            headers={'Content-Type': 'application/json'},
            data=json.dumps({'hello': 'world'}),
        )

    assert req_mock_response.status_code == expected_status_code
    assert req_mock_response.headers['Content-Type'] == expected_content_type
    assert req_mock_response.text == expected_data.decode()


def test_multiple_functions_same_path_different_type() -> None:
    """
    When multiple functions exist with the same path but have a different type,
    the mock matches them just the same.
    """
    app = Flask(__name__)

    @app.route('/<my_variable>')
    def _(_: float) -> str:
        ...  # pragma: no cover

    @app.route('/<int:my_variable>')
    def ___(my_variable: int) -> str:
        return 'Is int: ' + str(my_variable)

    @app.route('/<string:my_variable>')
    def ____(_: str) -> str:
        ...  # pragma: no cover

    test_client = app.test_client()
    response = test_client.get('/4')

    expected_status_code = 200
    expected_content_type = 'text/html; charset=utf-8'
    expected_data = b'Is int: 4'

    assert response.status_code == expected_status_code
    assert response.headers['Content-Type'] == expected_content_type
    assert response.data == expected_data

    with responses.RequestsMock(assert_all_requests_are_fired=False) as resp_m:
        add_flask_app_to_mock(
            mock_obj=resp_m,
            flask_app=app,
            base_url='http://www.example.com',
        )

        responses_response = requests.get('http://www.example.com/4')

    assert responses_response.status_code == expected_status_code
    assert responses_response.headers['Content-Type'] == expected_content_type
    assert responses_response.text == expected_data.decode()

    with requests_mock.Mocker() as resp_m:
        add_flask_app_to_mock(
            mock_obj=resp_m,
            flask_app=app,
            base_url='http://www.example.com',
        )

        req_mock_response = requests.get('http://www.example.com/4')

    assert req_mock_response.status_code == expected_status_code
    assert req_mock_response.headers['Content-Type'] == expected_content_type
    assert req_mock_response.text == expected_data.decode()


def test_query_string() -> None:
    """
    Query strings work.
    """
    app = Flask(__name__)

    @app.route('/')
    def _() -> str:
        result = request.args['frasier']
        return 'Hello: ' + str(result)

    test_client = app.test_client()
    response = test_client.get('/?frasier=crane')

    expected_status_code = 200
    expected_content_type = 'text/html; charset=utf-8'
    expected_data = b'Hello: crane'

    assert response.status_code == expected_status_code
    assert response.headers['Content-Type'] == expected_content_type
    assert response.data == expected_data

    with responses.RequestsMock(assert_all_requests_are_fired=False) as resp_m:
        add_flask_app_to_mock(
            mock_obj=resp_m,
            flask_app=app,
            base_url='http://www.example.com',
        )

        responses_response = requests.get(
            'http://www.example.com?frasier=crane',
        )

    assert responses_response.status_code == expected_status_code
    assert responses_response.headers['Content-Type'] == expected_content_type
    assert responses_response.text == expected_data.decode()

    with requests_mock.Mocker() as resp_m:
        add_flask_app_to_mock(
            mock_obj=resp_m,
            flask_app=app,
            base_url='http://www.example.com',
        )

        req_mock_response = requests.get(
            'http://www.example.com?frasier=crane',
        )

    assert req_mock_response.status_code == expected_status_code
    assert req_mock_response.headers['Content-Type'] == expected_content_type
    assert req_mock_response.text == expected_data.decode()


def test_cookies() -> None:
    """
    Cookies work.
    """
    app = Flask(__name__)

    @app.route('/', methods=['POST'])
    def _() -> Response:
        response = make_response()
        response.set_cookie('frasier_set', 'crane_set')
        assert request.cookies['frasier'] == 'crane'
        assert request.cookies['frasier2'] == 'crane2'
        response.data = 'Hello, World!'
        assert isinstance(response, Response)
        return response

    test_client = app.test_client()
    test_client.set_cookie(server_name='', key='frasier', value='crane')
    test_client.set_cookie(server_name='', key='frasier2', value='crane2')
    original_cookies = set(test_client.cookie_jar)
    response = test_client.post('/')

    expected_status_code = 200
    expected_content_type = 'text/html; charset=utf-8'
    expected_data = b'Hello, World!'

    (new_cookie,) = set(test_client.cookie_jar) - original_cookies
    assert new_cookie.name == 'frasier_set'
    assert new_cookie.value == 'crane_set'
    assert response.status_code == expected_status_code
    assert response.headers['Content-Type'] == expected_content_type
    assert response.data == expected_data

    with responses.RequestsMock(assert_all_requests_are_fired=False) as resp_m:
        add_flask_app_to_mock(
            mock_obj=resp_m,
            flask_app=app,
            base_url='http://www.example.com',
        )

        responses_response = requests.post(
            'http://www.example.com',
            cookies={
                'frasier': 'crane',
                'frasier2': 'crane2',
            },
        )

    assert responses_response.status_code == expected_status_code
    assert responses_response.headers['Content-Type'] == expected_content_type
    assert responses_response.text == expected_data.decode()
    assert responses_response.cookies['frasier_set'] == 'crane_set'

    with requests_mock.Mocker() as resp_m:
        add_flask_app_to_mock(
            mock_obj=resp_m,
            flask_app=app,
            base_url='http://www.example.com',
        )

        req_mock_response = requests.post(
            'http://www.example.com',
            cookies={
                'frasier': 'crane',
                'frasier2': 'crane2',
            },
        )

    assert req_mock_response.status_code == expected_status_code
    assert req_mock_response.headers['Content-Type'] == expected_content_type
    assert req_mock_response.text == expected_data.decode()
    assert req_mock_response.cookies['frasier_set'] == 'crane_set'
